@@ -1,4 +1,6 @@
 
+import inspect
+import io
 import os
 import sys
 import datetime
@@ -9,6 +11,9 @@ import re
 from os import path
 from six.moves import StringIO
 
+# use direct import to bypass freezegun
+from time import time
+
 from .unittest import TestResult, _TextTestResult, failfast
 
 
@@ -45,18 +50,7 @@ STDOUT_LINE = '\nStdout:\n%s'
 STDERR_LINE = '\nStderr:\n%s'
 
 
-def xml_safe_unicode(base, encoding='utf-8'):
-    """Return a unicode string containing only valid XML characters.
-
-    encoding - if base is a byte string it is first decoded to unicode
-        using this encoding.
-    """
-    if isinstance(base, six.binary_type):
-        base = base.decode(encoding)
-    return INVALID_XML_1_0_UNICODE_RE.sub('', base)
-
-
-def to_unicode(data):
+def _to_unicode(data):
     """Returns unicode in Python2 and str in Python3"""
     if six.PY3:
         return six.text_type(data)
@@ -67,8 +61,17 @@ def to_unicode(data):
         return repr(data).decode('utf8', 'replace')
 
 
-def safe_unicode(data, encoding=None):
-    return xml_safe_unicode(to_unicode(data), encoding)
+def safe_unicode(data, encoding='utf8'):
+    """Return a unicode string containing only valid XML characters.
+
+    encoding - if data is a byte string it is first decoded to unicode
+        using this encoding.
+    """
+    data = _to_unicode(data)
+    if isinstance(data, six.binary_type):
+        # e.g. IronPython, see #182
+        data = data.decode(encoding)
+    return INVALID_XML_1_0_UNICODE_RE.sub('', data)
 
 
 def testcase_name(test_method):
@@ -82,6 +85,46 @@ def testcase_name(test_method):
     return result
 
 
+class _DuplicateWriter(io.TextIOBase):
+    """
+    Duplicate output from the first handle to the second handle
+
+    The second handle is expected to be a StringIO and not to block.
+    """
+
+    def __init__(self, first, second):
+        super(_DuplicateWriter, self).__init__()
+        self._first = first
+        self._second = second
+
+    def flush(self):
+        self._first.flush()
+        self._second.flush()
+
+    def writable(self):
+        return True
+
+    def writelines(self, lines):
+        self._first.writelines(lines)
+        self._second.writelines(lines)
+
+    def write(self, b):
+        if isinstance(self._first, io.TextIOBase):
+            wrote = self._first.write(b)
+
+            if wrote is not None:
+                # expected to always succeed to write
+                self._second.write(b[:wrote])
+
+            return wrote
+        else:
+            # file-like object in Python2
+            # It doesn't return wrote bytes.
+            self._first.write(b)
+            self._second.write(b)
+            return len(b)
+
+
 class _TestInfo(object):
     """
     This class keeps useful information about the execution of a
@@ -91,12 +134,19 @@ class _TestInfo(object):
     # Possible test outcomes
     (SUCCESS, FAILURE, ERROR, SKIP) = range(4)
 
-    def __init__(self, test_result, test_method, outcome=SUCCESS, err=None, subTest=None):
+    OUTCOME_ELEMENTS = {
+        SUCCESS: None,
+        FAILURE: 'failure',
+        ERROR: 'error',
+        SKIP: 'skipped',
+    }
+
+    def __init__(self, test_result, test_method, outcome=SUCCESS, err=None, subTest=None, filename=None, lineno=None):
         self.test_result = test_result
         self.outcome = outcome
         self.elapsed_time = 0
         self.timestamp = datetime.datetime.min.replace(microsecond=0).isoformat()
-        if err:
+        if err is not None:
             if self.outcome != _TestInfo.SKIP:
                 self.test_exception_name = safe_unicode(err[0].__name__)
                 self.test_exception_message = safe_unicode(err[1])
@@ -115,8 +165,13 @@ class _TestInfo(object):
 
         self.test_name = testcase_name(test_method)
         self.test_id = test_method.id()
+
         if subTest:
             self.test_id = subTest.id()
+            self.test_description = self.test_result.getDescription(subTest)
+
+        self.filename = filename
+        self.lineno = lineno
 
     def id(self):
         return self.test_id
@@ -129,12 +184,6 @@ class _TestInfo(object):
         timestamp = datetime.datetime.fromtimestamp(self.test_result.stop_time)
         self.timestamp = timestamp.replace(microsecond=0).isoformat()
 
-    def get_description(self):
-        """
-        Return a text representation of the test method.
-        """
-        return self.test_description
-
     def get_error_info(self):
         """
         Return a text representation of an exception thrown by a test
@@ -152,13 +201,18 @@ class _XMLTestResult(_TextTestResult):
     def __init__(self, stream=sys.stderr, descriptions=1, verbosity=1,
                  elapsed_times=True, properties=None, infoclass=None):
         _TextTestResult.__init__(self, stream, descriptions, verbosity)
-        self.buffer = True  # we are capturing test output
         self._stdout_data = None
         self._stderr_data = None
+        self._stdout_capture = StringIO()
+        self.__stdout_saved = None
+        self._stderr_capture = StringIO()
+        self.__stderr_saved = None
         self.successes = []
         self.callback = None
         self.elapsed_times = elapsed_times
         self.properties = properties  # junit testsuite properties
+        self.filename = None
+        self.lineno = None
         if infoclass is None:
             self.infoclass = _TestInfo
         else:
@@ -170,6 +224,8 @@ class _XMLTestResult(_TextTestResult):
         Appends a `infoclass` to the given target list and sets a callback
         method to be called by stopTest method.
         """
+        test_info.filename = self.filename
+        test_info.lineno = self.lineno
         target_list.append(test_info)
 
         def callback():
@@ -189,27 +245,70 @@ class _XMLTestResult(_TextTestResult):
                 )
             elif self.dots:
                 self.stream.write(short_str)
+
+            self.stream.flush()
+
         self.callback = callback
 
     def startTest(self, test):
         """
         Called before execute each test method.
         """
-        self.start_time = time.time()
+        self.start_time = time()
         TestResult.startTest(self, test)
 
+        try:
+            if getattr(test, '_dt_test', None) is not None:
+                # doctest.DocTestCase
+                self.filename = test._dt_test.filename
+                self.lineno = test._dt_test.lineno
+            else:
+                # regular unittest.TestCase?
+                test_method = getattr(test, test._testMethodName)
+                test_class = type(test)
+                # Note: inspect can get confused with decorators, so use class.
+                self.filename = inspect.getsourcefile(test_class)
+                # Handle partial and partialmethod objects.
+                test_method = getattr(test_method, 'func', test_method)
+                _, self.lineno = inspect.getsourcelines(test_method)
+        except (AttributeError, TypeError):
+            # issue #188, #189, some frameworks can make test method opaque.
+            pass
+
         if self.showAll:
             self.stream.write('  ' + self.getDescription(test))
             self.stream.write(" ... ")
+            self.stream.flush()
+
+    def _setupStdout(self):
+        """
+        Capture stdout / stderr by replacing sys.stdout / sys.stderr
+        """
+        super(_XMLTestResult, self)._setupStdout()
+        self.__stdout_saved = sys.stdout
+        sys.stdout = _DuplicateWriter(sys.stdout, self._stdout_capture)
+        self.__stderr_saved = sys.stderr
+        sys.stderr = _DuplicateWriter(sys.stderr, self._stderr_capture)
+
+    def _restoreStdout(self):
+        """
+        Stop capturing stdout / stderr and recover sys.stdout / sys.stderr
+        """
+        if self.__stdout_saved:
+            sys.stdout = self.__stdout_saved
+            self.__stdout_saved = None
+        if self.__stderr_saved:
+            sys.stderr = self.__stderr_saved
+            self.__stderr_saved = None
+        self._stdout_capture.seek(0)
+        self._stdout_capture.truncate()
+        self._stderr_capture.seek(0)
+        self._stderr_capture.truncate()
+        super(_XMLTestResult, self)._restoreStdout()
 
     def _save_output_data(self):
-        # Only try to get sys.stdout and sys.sterr as they not be
-        # StringIO yet, e.g. when test fails during __call__
-        try:
-            self._stdout_data = sys.stdout.getvalue()
-            self._stderr_data = sys.stderr.getvalue()
-        except AttributeError:
-            pass
+        self._stdout_data = self._stdout_capture.getvalue()
+        self._stderr_data = self._stderr_capture.getvalue()
 
     def stopTest(self, test):
         """
@@ -220,7 +319,7 @@ class _XMLTestResult(_TextTestResult):
         # self._stderr_data = sys.stderr.getvalue()
 
         _TextTestResult.stopTest(self, test)
-        self.stop_time = time.time()
+        self.stop_time = time()
 
         if self.callback and callable(self.callback):
             self.callback()
@@ -232,7 +331,7 @@ class _XMLTestResult(_TextTestResult):
         """
         self._save_output_data()
         self._prepare_callback(
-            self.infoclass(self, test), self.successes, 'OK', '.'
+            self.infoclass(self, test), self.successes, 'ok', '.'
         )
 
     @failfast
@@ -268,14 +367,29 @@ class _XMLTestResult(_TextTestResult):
         Called when a subTest method raises an error.
         """
         if err is not None:
+
+            errorText = None
+            errorValue = None
+            errorList = None
+            if issubclass(err[0], test.failureException):
+                errorText = 'FAIL'
+                errorValue = self.infoclass.FAILURE
+                errorList = self.failures
+
+            else:
+                errorText = 'ERROR'
+                errorValue = self.infoclass.ERROR
+                errorList = self.errors
+
             self._save_output_data()
+
             testinfo = self.infoclass(
-                self, testcase, self.infoclass.ERROR, err, subTest=test)
-            self.errors.append((
+                self, testcase, errorValue, err, subTest=test)
+            errorList.append((
                 testinfo,
                 self._exc_info_to_string(err, testcase)
             ))
-            self._prepare_callback(testinfo, [], 'ERROR', 'E')
+            self._prepare_callback(testinfo, [], errorText, errorText[0])
 
     def addSkip(self, test, reason):
         """
@@ -284,8 +398,40 @@ class _XMLTestResult(_TextTestResult):
         self._save_output_data()
         testinfo = self.infoclass(
             self, test, self.infoclass.SKIP, reason)
+        testinfo.test_exception_name = 'skip'
+        testinfo.test_exception_message = reason
         self.skipped.append((testinfo, reason))
-        self._prepare_callback(testinfo, [], 'SKIP', 'S')
+        self._prepare_callback(testinfo, [], 'skip', 's')
+
+    def addExpectedFailure(self, test, err):
+        """
+        Missing in xmlrunner, copy-pasted from xmlrunner addError.
+        """
+        self._save_output_data()
+
+        testinfo = self.infoclass(self, test, self.infoclass.SKIP, err)
+        testinfo.test_exception_name = 'XFAIL'
+        testinfo.test_exception_message = 'expected failure: {}'.format(testinfo.test_exception_message)
+
+        self.expectedFailures.append((testinfo, self._exc_info_to_string(err, test)))
+        self._prepare_callback(testinfo, [], 'expected failure', 'x')
+
+    @failfast
+    def addUnexpectedSuccess(self, test):
+        """
+        Missing in xmlrunner, copy-pasted from xmlrunner addSuccess.
+        """
+        self._save_output_data()
+
+        testinfo = self.infoclass(self, test)  # do not set outcome here because it will need exception
+        testinfo.outcome = self.infoclass.ERROR
+        # But since we want to have error outcome, we need to provide additional fields:
+        testinfo.test_exception_name = 'UnexpectedSuccess'
+        testinfo.test_exception_message = ('Unexpected success: This test was marked as expected failure but passed, '
+                                           'please review it')
+
+        self.unexpectedSuccesses.append((testinfo, 'unexpected success'))
+        self._prepare_callback(testinfo, [], 'unexpected success', 'u')
 
     def printErrorList(self, flavour, errors):
         """
@@ -295,10 +441,11 @@ class _XMLTestResult(_TextTestResult):
             self.stream.writeln(self.separator1)
             self.stream.writeln(
                 '%s [%.3fs]: %s' % (flavour, test_info.elapsed_time,
-                                    test_info.get_description())
+                                    test_info.test_description)
             )
             self.stream.writeln(self.separator2)
             self.stream.writeln('%s' % test_info.get_error_info())
+            self.stream.flush()
 
     def _get_info_by_testcase(self):
         """
@@ -309,7 +456,7 @@ class _XMLTestResult(_TextTestResult):
         tests_by_testcase = {}
 
         for tests in (self.successes, self.failures, self.errors,
-                      self.skipped):
+                      self.skipped, self.expectedFailures, self.unexpectedSuccesses):
             for test_info in tests:
                 if isinstance(test_info, tuple):
                     # This is a skipped, error or a failure test case
@@ -340,9 +487,12 @@ class _XMLTestResult(_TextTestResult):
         """
         testsuite = xml_document.createElement('testsuite')
         parentElement.appendChild(testsuite)
+        module_name = suite_name.rpartition('.')[0]
+        file_name = module_name.replace('.', '/') + '.py'
 
         testsuite.setAttribute('name', suite_name)
         testsuite.setAttribute('tests', str(len(tests)))
+        testsuite.setAttribute('file', file_name)
 
         testsuite.setAttribute(
             'time', '%.3f' % sum(map(lambda e: e.elapsed_time, tests))
@@ -366,28 +516,6 @@ class _XMLTestResult(_TextTestResult):
         for test in tests:
             _XMLTestResult._report_testcase(test, testsuite, xml_document)
 
-        systemout = xml_document.createElement('system-out')
-        testsuite.appendChild(systemout)
-
-        stdout = StringIO()
-        for test in tests:
-            # Merge the stdout from the tests in a class
-            if test.stdout is not None:
-                stdout.write(test.stdout)
-        _XMLTestResult._createCDATAsections(
-            xml_document, systemout, stdout.getvalue())
-
-        systemerr = xml_document.createElement('system-err')
-        testsuite.appendChild(systemerr)
-
-        stderr = StringIO()
-        for test in tests:
-            # Merge the stderr from the tests in a class
-            if test.stderr is not None:
-                stderr.write(test.stderr)
-        _XMLTestResult._createCDATAsections(
-            xml_document, systemerr, stderr.getvalue())
-
         return testsuite
 
     _report_testsuite = staticmethod(_report_testsuite)
@@ -430,25 +558,45 @@ class _XMLTestResult(_TextTestResult):
         testcase.setAttribute('time', '%.3f' % test_result.elapsed_time)
         testcase.setAttribute('timestamp', test_result.timestamp)
 
-        if (test_result.outcome != test_result.SUCCESS):
-            elem_name = ('failure', 'error', 'skipped')[test_result.outcome-1]
-            failure = xml_document.createElement(elem_name)
-            testcase.appendChild(failure)
-            if test_result.outcome != test_result.SKIP:
-                failure.setAttribute(
-                    'type',
-                    test_result.test_exception_name
-                )
-                failure.setAttribute(
-                    'message',
-                    test_result.test_exception_message
-                )
+        if test_result.filename is not None:
+            # Try to make filename relative to current directory.
+            filename = os.path.relpath(test_result.filename)
+            filename = test_result.filename if filename.startswith('../') else filename
+            testcase.setAttribute('file', filename)
+
+        if test_result.lineno is not None:
+            testcase.setAttribute('line', str(test_result.lineno))
+
+        result_elem_name = test_result.OUTCOME_ELEMENTS[test_result.outcome]
+
+        if result_elem_name is not None:
+            result_elem = xml_document.createElement(result_elem_name)
+            testcase.appendChild(result_elem)
+
+            result_elem.setAttribute(
+                'type',
+                test_result.test_exception_name
+            )
+            result_elem.setAttribute(
+                'message',
+                test_result.test_exception_message
+            )
+            if test_result.get_error_info():
                 error_info = safe_unicode(test_result.get_error_info())
                 _XMLTestResult._createCDATAsections(
-                    xml_document, failure, error_info)
-            else:
-                failure.setAttribute('type', 'skip')
-                failure.setAttribute('message', test_result.test_exception_message)
+                    xml_document, result_elem, error_info)
+
+        if test_result.stdout is not None:
+            systemout = xml_document.createElement('system-out')
+            testcase.appendChild(systemout)
+            _XMLTestResult._createCDATAsections(
+                xml_document, systemout, test_result.stdout)
+
+        if test_result.stderr is not None:
+            systemout = xml_document.createElement('system-err')
+            testcase.appendChild(systemout)
+            _XMLTestResult._createCDATAsections(
+                xml_document, systemout, test_result.stderr)
 
     _report_testcase = staticmethod(_report_testcase)
 
@@ -485,20 +633,27 @@ class _XMLTestResult(_TextTestResult):
             testsuite = _XMLTestResult._report_testsuite(
                 suite_name, tests, doc, parentElement, self.properties
             )
-            xml_content = doc.toprettyxml(
-                indent='\t',
-                encoding=test_runner.encoding
-            )
 
             if outputHandledAsString:
+                xml_content = doc.toprettyxml(
+                    indent='\t',
+                    encoding=test_runner.encoding
+                )
                 filename = path.join(
                     test_runner.output,
                     'TEST-%s.xml' % suite_name)
                 with open(filename, 'wb') as report_file:
                     report_file.write(xml_content)
 
+                if self.showAll:
+                    self.stream.writeln('Generated XML report: {}'.format(filename))
+
         if not outputHandledAsString:
             # Assume that test_runner.output is a stream
+            xml_content = doc.toprettyxml(
+                indent='\t',
+                encoding=test_runner.encoding
+            )
             test_runner.output.write(xml_content)
 
     def _exc_info_to_string(self, err, test):
@@ -526,16 +681,8 @@ class _XMLTestResult(_TextTestResult):
             msgLines = traceback.format_exception(exctype, value, tb)
 
         if self.buffer:
-            # Only try to get sys.stdout and sys.sterr as they not be
-            # StringIO yet, e.g. when test fails during __call__
-            try:
-                output = sys.stdout.getvalue()
-            except AttributeError:
-                output = None
-            try:
-                error = sys.stderr.getvalue()
-            except AttributeError:
-                error = None
+            output = self._stdout_capture.getvalue()
+            error = self._stdout_capture.getvalue()
             if output:
                 if not output.endswith('\n'):
                     output += '\n'
